;; Generated package description from diff-hl.el
(define-package "diff-hl" "1.8.1" "Highlight uncommitted changes" '((cl-lib "0.2")) :url "https://github.com/dgutov/diff-hl" :keywords '("vc" "diff"))
